"""
BoseDiscovery Module

This module provides functionality to discover Bose devices on the local network using Zeroconf.
It scans for devices broadcasting the "_bose-passport._tcp.local." service and extracts key
information such as the device GUID and IP address.
"""

from zeroconf import Zeroconf, ServiceBrowser, ServiceStateChange
from typing import List, Dict
import time


class BoseDiscovery:
    """
    Discover Bose devices on the local network using Zeroconf.

    This class creates a Zeroconf instance (if not provided) and listens for devices
    advertising the "_bose-passport._tcp.local." service. When a service is added, it resolves
    the service to extract the device's GUID and IP address.

    Attributes:
        zeroconf (Zeroconf): The Zeroconf instance used for network discovery.
        devices (List[Dict[str, str]]): A list of discovered devices with keys "GUID" and "IP".
    """

    def __init__(self, zeroconf: Zeroconf = None) -> None:
        """
        Initialize a BoseDiscovery instance.

        If no Zeroconf instance is provided, a new one is created.

        Args:
            zeroconf (Optional[Zeroconf]): An optional Zeroconf instance to use.
        """
        if zeroconf is None:
            zeroconf = Zeroconf()
        self.zeroconf = zeroconf
        self.devices: List[Dict[str, str]] = []

    def _on_service_state_change(self, zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange) -> None:
        """
        Callback invoked on service state changes.

        When a new service is added, this method calls the _resolve_service method
        to process the service details.

        Args:
            zeroconf (Zeroconf): The Zeroconf instance triggering the callback.
            service_type (str): The type of service that changed.
            name (str): The name of the service.
            state_change (ServiceStateChange): The state change (e.g., Added, Removed).
        """
        if state_change == ServiceStateChange.Added:
            self._resolve_service(name)

    def _resolve_service(self, name: str) -> None:
        """
        Resolve the details of a discovered service.

        Retrieves service information for the "_bose-passport._tcp.local." service using the provided name.
        If the service contains a "GUID" property and at least one IP address, it appends a dictionary with
        the GUID and the first IP address to the devices list.

        Args:
            name (str): The name of the service to resolve.
        """
        info = self.zeroconf.get_service_info("_bose-passport._tcp.local.", name)
        if info:
            guid = info.properties.get(b"GUID")
            if guid:
                guid = guid.decode("utf-8")
            addresses = [addr for addr in info.parsed_addresses()]
            if addresses:
                self.devices.append({"GUID": guid, "IP": addresses[0]})

    def discover_devices(self, timeout: int = 5) -> List[Dict[str, str]]:
        """
        Discover Bose devices advertising the "_bose-passport._tcp.local." service.

        This method initializes the devices list, creates a ServiceBrowser to listen for
        service events, waits for a given timeout period, and then closes the Zeroconf instance.

        Args:
            timeout (int): Time in seconds to wait for service discovery. Defaults to 5.

        Returns:
            List[Dict[str, str]]: A list of dictionaries, each containing the keys "GUID" and "IP"
            for a discovered device.
        """
        self.devices = []

        # Create a ServiceBrowser that listens for service events and calls _on_service_state_change.
        listener = ServiceBrowser(self.zeroconf, "_bose-passport._tcp.local.", handlers=[self._on_service_state_change])

        try:
            time.sleep(timeout)
        finally:
            self.zeroconf.close()

        return self.devices


"""
Bose Speaker control using its WebSocket (like the BOSE app)

In order to control the device locally, you need to obtain the control token and device ID.
The control token is acquired from the online BOSE API (using BoseAuth.py), while the device ID
is obtained by discovering the device on the local network (using BoseDiscovery.py).

Note: The token is only valid for a limited time and does not renew automatically.
You may need to refetch the token periodically.
"""

from __future__ import annotations
import json
import asyncio
import logging
from ssl import SSLContext, CERT_NONE, PROTOCOL_TLS_CLIENT
import websockets
from threading import Event
from typing import Any, Callable, Dict, List, Optional, Union
from pybose.BoseAuth import BoseAuth

from . import BoseResponse as BR

# Default resources subscribed when connecting (same as in the BOSE app)
DEFAULT_SUBSCRIBE_RESOURCES: List[str] = [
    "/bluetooth/sink/list",
    "/system/power/control",
    "/audio/avSync",
    "/bluetooth/source/list",
    "/audio/bass",
    "/device/assumed/TVs",
    "/network/wifi/siteScan",
    "/system/update/start",
    "/system/update/status",
    "/bluetooth/sink/macAddr",
    "/content/nowPlaying/rating",
    "/bluetooth/source/stopScan",
    "/system/setup",
    "/homekit/info",
    "/bluetooth/source/pairStatus",
    "/device/configuredDevices",
    "/bluetooth/source/status",
    "/cast/teardown",
    "/bluetooth/sink/status",
    "/cast/setup",
    "/cec",
    "/cloudSync",
    "/system/challenge",
    "/bluetooth/sink/remove",
    "/bluetooth/source/connect",
    "/remote/integration/brandList",
    "/subscription",
    "/network/status",
    "/bluetooth/source/scanResult",
    "/content/playbackRequest",
    "/audio/eqSelect",
    "/audio/height",
    "/content/transportControl",
    "/grouping/activeGroups",
    "/audio/mode",
    "/bluetooth/source/pair",
    "/bluetooth/source/capability",
    "/bluetooth/source/disconnect",
    "/audio/subwooferGain",
    "/voice/setup/start",
    "/audio/center",
    "/network/wifi/status",
    "/content/nowPlaying/repeat",
    "/system/sources",
    "/content/nowPlaying",
    "/system/power/macro",
    "/bluetooth/sink/pairable",
    "/network/wifi/profile",
    "/cast/settings",
    "/audio/zone",
    "/content/nowPlaying/shuffle",
    "/bluetooth/source/capabilitySettings",
    "/remote/integration",
    "/audio/surround",
    "/accessories",
    "/audio/treble",
    "/adaptiq",
    "/accessories/playTones",
    "/system/power/timeouts",
    "/audio/dualMonoSelect",
    "/system/info",
    "/system/sources/status",
    "/audio/rebroadcastLatency/mode",
    "/audio/format",
    "/bluetooth/source/connectionStatus",
    "/system/power/mode/opticalAutoWake",
    "/content/nowPlaying/favorite",
    "/system/productSettings",
    "/bluetooth/sink/connectionStatus",
    "/bluetooth/source/remove",
    "/audio/autoVolume",
    "/system/capabilities",
    "/audio/volume/increment",
    "/bluetooth/sink/connect",
    "/bluetooth/source/volume",
    "/bluetooth/sink/disconnect",
    "/system/reset",
    "/audio/volume/decrement",
    "/audio/volume",
    "/remote/integration/directEntry",
    "/device/configure",
    "/device/setup",
    "/bluetooth/source/scan",
    "/voice/settings",
    "/system/activated",
]


class BoseSpeaker:
    def __init__(
        self,
        control_token: str,
        host: str,
        device_id: Optional[str] = None,
        version: int = 1,
        auto_reconnect: bool = True,
        bose_auth: BoseAuth = None,
    ) -> None:
        self._control_token: str = control_token
        self._device_id: Optional[str] = device_id
        self._host: str = host
        self._version: int = version
        self._websocket: Optional[websockets.connect] = None
        self._ssl_context: SSLContext = SSLContext(PROTOCOL_TLS_CLIENT)
        self._ssl_context.check_hostname = False
        self._ssl_context.verify_mode = CERT_NONE
        self._subprotocol: str = "eco2"
        self._req_id: int = 1
        self._url: str = f"wss://{self._host}:8082/?product=Madrid-iOS:31019F02-F01F-4E73-B495-B96D33AD3664"
        self._responses: List[BR.BoseMessage] = []
        self._stop_event: Event = Event()
        self._receiver_task: Optional[asyncio.Task] = None
        self._receivers: Dict[int, Callable[[BR.BoseMessage], None]] = {}
        self._subscribed_resources: List[str] = []
        self._message_queue: asyncio.Queue = asyncio.Queue()
        self._capabilities: Optional[BR.Capabilities] = None
        self._auto_reconnect = auto_reconnect
        self._bose_auth: BoseAuth = bose_auth

    async def connect(self) -> None:
        """Connect to the WebSocket and start the receiver loop."""
        await self.disconnect()
        self._websocket = await websockets.connect(
            self._url, subprotocols=[self._subprotocol], ssl=self._ssl_context
        )
        logging.info("WebSocket connection established.")
        self._stop_event.clear()
        self._receiver_task = asyncio.create_task(self._receiver_loop())
        if self._subscribed_resources:
            logging.debug("Subscribing to resources from previous session.")
            await self.subscribe(self._subscribed_resources)
        await self.get_capabilities()

    async def disconnect(self) -> None:
        """Stop the receiver loop and close the WebSocket connection."""
        self._stop_event.set()
        if self._receiver_task:
            await self._receiver_task
        if self._websocket:
            await self._websocket.close()
        logging.info("WebSocket connection closed.")

    def attach_receiver(self, callback: Callable[[BR.BoseMessage], None]) -> int:
        """Attach a callback to receive unsolicited messages."""
        receiver_id = max(self._receivers.keys(), default=0) + 1
        self._receivers[receiver_id] = callback
        return receiver_id

    def detach_receiver(self, receiver_id: int) -> None:
        """Detach a previously attached receiver."""
        self._receivers.pop(receiver_id, None)

    async def _request(
        self,
        resource: str,
        method: str,
        body: Optional[Dict[str, Any]] = None,
        withHeaders: bool = False,
        waitForResponse: bool = True,
        version: Optional[int] = None,
        checkCapabilities: bool = True,
    ) -> Dict[str, Any]:
        """Send a request over the WebSocket and wait for the matching response."""
        if body is None:
            body = {}

        if self._bose_auth:
            if not self._bose_auth.is_token_valid():
                logging.warning("Token is not valid. Refreshing token.")
                control_token = await self._bose_auth.do_token_refresh()
                self._control_token = control_token.get("access_token")

        token: str = self._control_token
        req_id: int = self._req_id
        self._req_id += 1

        version = version if version is not None else self._version

        if checkCapabilities and not self.has_capability(resource):
            raise BoseFunctionNotSupportedException(
                f"Resource {resource} is not supported by the device."
            )

        header: BR.BoseHeader = {
            "device": self._device_id,
            "method": method,
            "msgtype": "REQUEST",
            "reqID": req_id,
            "resource": resource,
            "status": 200,
            "token": token,
            "version": version,
        }
        message: BR.BoseMessage = {"body": body, "header": header}

        if self._device_id is None:
            await self._message_queue.put(message)
            logging.debug(
                f"Waiting for deviceID. Queued message: {json.dumps(message, indent=4)}"
            )
        else:
            try:
                await self._websocket.send(json.dumps(message))
            except websockets.ConnectionClosed:
                if self._auto_reconnect:
                    logging.warning(
                        "WebSocket connection is closed. Reconnecting before sending message."
                    )
                    await self.connect()
                    await self._websocket.send(json.dumps(message))
                else:
                    logging.error(
                        "WebSocket connection is closed. Cannot send message."
                    )
                    raise Exception("WebSocket connection is closed.")

            logging.debug(f"Sent message: {json.dumps(message, indent=4)}")

        if not waitForResponse:
            return {}

        while True:
            for response in self._responses:
                resp_header = response.get("header", {})
                if (
                    resp_header.get("msgtype") == BR.BoseHeaderMsgTypeEnum.RESPONSE
                    and resp_header.get("reqID") == req_id
                ):
                    self._responses.remove(response)
                    status = resp_header.get("status")
                    if status is None:
                        raise BoseRequestException(
                            method,
                            resource,
                            body,
                            999,
                            999,
                            f"pybose could not parse the message: {response}",
                        )
                    if status != 200:
                        error = response.get(
                            "error",
                            {
                                "status": 999,
                                "message": f"pybose could not determine the error, but status code is {status}",
                            },
                        )
                        raise BoseRequestException(
                            method,
                            resource,
                            body,
                            status,
                            error["code"],
                            error["message"],
                        )
                    return response["body"] if not withHeaders else response
            await asyncio.sleep(0.1)

    async def _receiver_loop(self) -> None:
        """Continuously receive and process messages from the WebSocket."""
        try:
            while not self._stop_event.is_set():
                message = await self._websocket.recv()
                logging.debug(f"Received message: {message}")
                parsed_message: BR.BoseMessage = json.loads(message)
                header: BR.BoseHeader = parsed_message.get("header", {})
                if header.get("device") is not None and self._device_id is None:
                    self._device_id = header["device"]
                    logging.debug(
                        f"Received first message from device. Device ID: {self._device_id}"
                    )
                    logging.debug(
                        f"Sending {self._message_queue.qsize()} queued messages."
                    )
                    while not self._message_queue.empty():
                        queued_msg = await self._message_queue.get()
                        queued_msg["header"]["device"] = self._device_id
                        await self._websocket.send(json.dumps(queued_msg))
                if (
                    header.get("msgtype") == BR.BoseHeaderMsgTypeEnum.RESPONSE
                    and header.get("reqID") is not None
                ):
                    logging.debug(f"Response received for reqID: {header['reqID']}")
                    self._responses.append(parsed_message)
                else:
                    for receiver in self._receivers.values():
                        receiver(parsed_message)
        except websockets.ConnectionClosed:
            logging.warning("WebSocket connection lost.")
            if self._auto_reconnect:
                logging.info("Reconnecting...")
                await self.connect()
            else:
                logging.error("WebSocket connection closed.")
        except Exception as e:
            if not self._stop_event.is_set():
                logging.error(f"Error in receiver loop: {e}")

    async def get_capabilities(self) -> BR.Capabilities:
        """Retrieve the device capabilities."""
        self._capabilities = BR.Capabilities(
            await self._request("/system/capabilities", "GET", checkCapabilities=False)
        )
        return self._capabilities

    def has_capability(self, endpoint: str) -> bool:
        """Return True if the device has the specified capability."""
        if self._capabilities is None:
            raise BoseCapabilitiesNotLoadedException()
        groups: List[BR.CapabilityGroup] = self._capabilities.get("group", [])
        endpoints: List[str] = [
            ep.get("endpoint") for group in groups for ep in group.get("endpoints", [])
        ]
        return endpoint in endpoints

    async def get_system_info(self) -> BR.SystemInfo:
        """Retrieve system information."""
        return BR.SystemInfo(await self._request("/system/info", "GET"))

    async def get_audio_volume(self) -> BR.AudioVolume:
        """Retrieve the current audio volume."""
        return BR.AudioVolume(await self._request("/audio/volume", "GET"))

    async def set_audio_volume(self, volume: int) -> BR.AudioVolume:
        """Set the audio volume to the specified value."""
        body = {"value": volume}
        return BR.AudioVolume(await self._request("/audio/volume", "PUT", body))

    async def get_now_playing(self) -> BR.ContentNowPlaying:
        """Retrieve the currently playing content."""
        return BR.ContentNowPlaying(await self._request("/content/nowPlaying", "GET"))

    async def get_bluetooth_status(self) -> dict:
        """Retrieve Bluetooth status."""
        return await self._request("/bluetooth/source/status", "GET")

    async def get_power_state(self) -> BR.SystemPowerControl:
        """Retrieve the power state of the device."""
        return await self._request("/system/power/control", "GET")

    async def set_power_state(self, state: bool) -> None:
        """Set the device power state."""
        body = {"power": "ON" if state else "OFF"}
        await self._request("/system/power/control", "POST", body)

    async def _control_transport(self, control: str) -> BR.ContentNowPlaying:
        """Send a transport control command."""
        body = {"state": control}
        return BR.ContentNowPlaying(
            await self._request("/content/transportControl", "PUT", body)
        )

    async def pause(self) -> BR.ContentNowPlaying:
        """Pause playback."""
        return await self._control_transport("PAUSE")

    async def play(self) -> BR.ContentNowPlaying:
        """Resume playback."""
        return await self._control_transport("PLAY")

    async def skip_next(self) -> BR.ContentNowPlaying:
        """Skip to the next item."""
        return await self._control_transport("SKIPNEXT")

    async def skip_previous(self) -> BR.ContentNowPlaying:
        """Skip to the previous item."""
        return await self._control_transport("SKIPPREVIOUS")

    async def seek(self, position: Union[float, int]) -> BR.ContentNowPlaying:
        """Seek to a specified position (in seconds)."""
        body = {"position": position, "state": "SEEK"}
        return await self._request("/content/transportControl", "PUT", body)

    async def request_playback_preset(
        self, preset: BR.Preset, initiator_id: str
    ) -> bool:
        """Request a playback preset."""
        content_item = preset["actions"][0]["payload"]["contentItem"]
        body = {
            "source": content_item.get("source"),
            "initiatorID": initiator_id,
            "sourceAccount": content_item.get("sourceAccount"),
            "preset": {
                "location": content_item.get("location"),
                "name": content_item.get("name"),
                "containerArt": content_item.get("containerArt"),
                "presetable": content_item.get("presetable"),
                "type": content_item.get("type"),
            },
        }
        return await self._request("/content/playbackRequest", "POST", body)

    def get_device_id(self) -> Optional[str]:
        """Return the device ID (if available)."""
        return self._device_id

    async def subscribe(
        self, resources: List[str] = DEFAULT_SUBSCRIBE_RESOURCES
    ) -> dict:
        """Subscribe to a list of resources."""
        body = {
            "notifications": [
                {"resource": resource, "version": 1} for resource in resources
            ]
        }
        self._subscribed_resources = resources
        return await self._request("/subscription", "PUT", body, version=2)

    async def switch_tv_source(self) -> BR.ContentNowPlaying:
        """Switch the speaker’s source to TV."""
        return await self.set_source("PRODUCT", "TV")

    async def set_source(self, source: str, sourceAccount: str) -> BR.ContentNowPlaying:
        """Set the playback source."""
        body = {"source": source, "sourceAccount": sourceAccount}
        return BR.ContentNowPlaying(
            await self._request("/content/playbackRequest", "POST", body)
        )

    async def get_sources(self) -> BR.Sources:
        """Retrieve available sources."""
        return BR.Sources(await self._request("/system/sources", "GET"))

    async def get_audio_setting(self, option: str) -> BR.Audio:
        """Retrieve an audio setting value."""
        if option not in [
            "bass",
            "treble",
            "center",
            "subwooferGain",
            "height",
            "avSync",
        ]:
            raise BoseInvalidAudioSettingException(option)
        return BR.Audio(await self._request("/audio/" + option, "GET"))

    async def set_audio_setting(self, option: str, value: int) -> BR.Audio:
        """Set an audio setting value."""
        if option not in [
            "bass",
            "treble",
            "center",
            "subwooferGain",
            "height",
            "avSync",
        ]:
            raise BoseInvalidAudioSettingException(option)
        return BR.Audio(
            await self._request("/audio/" + option, "POST", {"value": value})
        )

    async def get_accessories(self) -> BR.Accessories:
        """Retrieve accessories information."""
        return BR.Accessories(await self._request("/accessories", "GET"))

    async def put_accessories(
        self, subs_enabled: Optional[bool] = None, rears_enabled: Optional[bool] = None
    ) -> bool:
        """Update accessories settings."""
        if subs_enabled is None and rears_enabled is None:
            accessories = await self.get_accessories()
            subs_enabled = accessories.enabled.subs  # type: ignore
            rears_enabled = accessories.enabled.rears  # type: ignore
        body = {"enabled": {"rears": rears_enabled, "subs": subs_enabled}}
        return await self._request("/accessories", "PUT", body)

    async def get_battery_status(self) -> BR.Battery:
        """Retrieve battery status."""
        return BR.Battery(await self._request("/system/battery", "GET"))

    async def get_audio_mode(self) -> BR.AudioMode:
        """Retrieve the audio mode."""
        return BR.AudioMode(await self._request("/audio/mode", "GET"))

    async def set_audio_mode(self, mode: str) -> bool:
        """Set the audio mode."""
        result = await self._request("/audio/mode", "POST", {"value": mode})
        return result.get("value") == mode

    async def get_dual_mono_setting(self) -> BR.DualMonoSettings:
        """Retrieve the dual mono setting."""
        return BR.DualMonoSettings(await self._request("/audio/dualMonoSelect", "GET"))

    async def set_dual_mono_setting(self, value: Union[int, str]) -> bool:
        """Set the dual mono setting."""
        result = await self._request("/audio/dualMonoSelect", "POST", {"value": value})
        return result.get("value") == value

    async def get_rebroadcast_latency_mode(self) -> BR.RebroadcastLatencyMode:
        """Retrieve the rebroadcast latency mode."""
        return await self._request("/audio/rebroadcastLatency/mode", "GET")

    async def set_rebroadcast_latency_mode(self, mode: str) -> bool:
        """Set the rebroadcast latency mode."""
        result = await self._request(
            "/audio/rebroadcastLatency/mode", "PUT", {"mode": mode}
        )
        return result.get("value") == mode

    async def get_active_groups(self) -> List[BR.ActiveGroup]:
        """Retrieve active groups."""
        groups = await self._request("/grouping/activeGroups", "GET")
        return [BR.ActiveGroup(group) for group in groups.get("activeGroups", [])]

    async def set_active_group(self, other_product_ids: List[str]) -> bool:
        """Set the active group including this device and other products."""
        body = {"products": [{"productId": self._device_id, "role": "NORMAL"}]}
        for product_id in other_product_ids:
            body["products"].append({"productId": product_id, "role": "NORMAL"})
        return await self._request("/grouping/activeGroups", "POST", body)

    async def add_to_active_group(
        self, active_group_id: str, other_product_ids: List[str]
    ) -> bool:
        """Add products to the active group."""
        body = {
            "activeGroupId": active_group_id,
            "addProducts": [
                {"productId": pid, "role": "NORMAL"} for pid in other_product_ids
            ],
            "addGroups": [],
            "removeGroups": [],
            "removeProducts": [],
        }
        return await self._request("/grouping/activeGroups", "PUT", body)

    async def remove_from_active_group(
        self, active_group_id: str, other_product_ids: List[str]
    ) -> bool:
        """Remove products from the active group."""
        body = {
            "name": "",
            "activeGroupId": active_group_id,
            "addProducts": [],
            "addGroups": [],
            "removeGroups": [],
            "removeProducts": [
                {"productId": pid, "role": "NORMAL"} for pid in other_product_ids
            ],
        }
        return await self._request("/grouping/activeGroups", "PUT", body)

    async def stop_active_groups(self) -> bool:
        """Stop all active groups."""
        return await self._request("/grouping/activeGroups", "DELETE")

    async def get_system_timeout(self) -> BR.SystemTimeout:
        """Retrieve the system timeout settings."""
        return BR.SystemTimeout(await self._request("/system/power/timeouts", "GET"))

    async def set_system_timeout(
        self, no_audio: bool, no_video: bool
    ) -> BR.SystemTimeout:
        """Set system timeout settings."""
        body = {"noAudio": no_audio, "noVideo": no_video}
        return BR.SystemTimeout(
            await self._request("/system/power/timeouts", "PUT", body)
        )

    async def get_cec_settings(self) -> BR.CecSettings:
        """Retrieve the CEC settings."""
        return BR.CecSettings(await self._request("/cec", "GET"))

    async def set_cec_settings(
        self, mode: BR.CecSettingsSupportedValuesEnum
    ) -> BR.CecSettings:
        """Set the CEC settings."""
        return BR.CecSettings(await self._request("/cec", "PUT", {"mode": mode}))

    async def get_product_settings(self) -> BR.ProductSettings:
        """Retrieve the product settings."""
        return BR.ProductSettings(await self._request("/system/productSettings", "GET"))

    async def get_network_status(self) -> BR.NetworkStatus:
        """Retrieve the network status."""
        return BR.NetworkStatus(await self._request("/network/status", "GET"))


class BoseFunctionNotSupportedException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class BoseCapabilitiesNotLoadedException(Exception):
    def __init__(self) -> None:
        self.message = "Capabilities not loaded yet."
        super().__init__(self.message)


class BoseInvalidAudioSettingException(Exception):
    def __init__(self, setting: str) -> None:
        self.message = f"Invalid audio setting: {setting}"
        super().__init__(self.message)


class BoseRequestException(Exception):
    def __init__(
        self,
        method: str,
        resource: str,
        body: Dict[str, Any],
        http_status: int,
        error_status: int,
        message: str,
    ) -> None:
        self.message = message
        super().__init__(
            f"'{method} {resource}' returned {http_status}: Bose Error #{error_status} - {self.message}"
        )
        logging.debug(f"Request body for previous error: {json.dumps(body, indent=4)}")



from typing import TypedDict, List, Optional, Dict


class BoseHeaderMsgTypeEnum(enumerate):
    RESPONSE = "RESPONSE"
    NOTIFY = "NOTIFY"


class BoseHeader(TypedDict):
    device: str
    resource: str
    method: str
    msgtype: BoseHeaderMsgTypeEnum
    reqID: int
    version: float
    status: int
    token: str
    targetGuid: str


class BoseMessage(TypedDict):
    header: BoseHeader
    body: dict


# Capabilities
class CapabilityEndpoint(TypedDict):
    endpoint: str


class CapabilityGroup(TypedDict):
    apiGroup: str
    endpoints: List[CapabilityEndpoint]
    version: float


class Capabilities(TypedDict):
    group: list[CapabilityGroup]


# SystemInfo
class SystemInfo(TypedDict):
    countryCode: str
    defaultName: str
    limitedFeatures: bool
    name: str
    productColor: int
    productId: int
    productName: str
    productType: str
    regionCode: str
    serialNumber: str
    softwareVersion: str
    variantId: int


# AudioVolume
class VolumeProperties(TypedDict):
    maxLimit: int
    maxLimitOverride: bool
    minLimit: int
    startupVolume: int
    startupVolumeOverride: bool


class AudioVolume(TypedDict):
    defaultOn: int
    max: int
    min: int
    muted: bool
    properties: VolumeProperties
    value: int


# ContentNowPlaying
class ContentItem(TypedDict, total=False):
    isLocal: bool
    presetable: bool
    source: str
    sourceAccount: str
    containerArt: str


class ContainerCapabilities(TypedDict, total=False):
    favoriteSupported: bool
    ratingsSupported: bool
    repeatSupported: bool
    resumeSupported: bool
    seekRelativeBackwardSupported: bool
    seekRelativeForwardSupported: bool
    shuffleSupported: bool
    skipNextSupported: bool
    skipPreviousSupported: bool


class Container(TypedDict, total=False):
    contentItem: Optional[ContentItem]
    capabilities: Optional[ContainerCapabilities]


class Source(TypedDict, total=False):
    sourceDisplayName: str
    sourceID: str


class Metadata(TypedDict, total=False):
    album: str
    artist: str
    duration: int
    trackName: str


class State(TypedDict, total=False):
    canFavorite: bool
    canPause: bool
    canRate: bool
    canRepeat: bool
    canSeek: bool
    canShuffle: bool
    canSkipNext: bool
    canSkipPrevious: bool
    canStop: bool
    quality: str
    repeat: str
    shuffle: str
    status: str
    timeIntoTrack: int
    timestamp: str


class Track(TypedDict, total=False):
    contentItem: Optional[ContentItem]
    favorite: str
    rating: str


class ContentNowPlaying(TypedDict, total=False):
    collectData: bool
    container: Optional[Container]
    source: Optional[Source]
    initiatorID: str
    metadata: Optional[Metadata]
    state: Optional[State]
    track: Optional[Track]


# System Power Control
class SystemPowerControl(TypedDict):
    power: str


# Sources
class SourceData(TypedDict, total=False):
    accountId: str
    displayName: str
    local: bool
    multiroom: bool
    sourceAccountName: str
    sourceName: str
    status: str
    visible: bool


class SourceProperties(TypedDict, total=False):
    supportedActivationKeys: List[str]
    supportedDeviceTypes: List[str]
    supportedFriendlyNames: List[str]
    supportedInputRoutes: List[str]


class Sources(TypedDict):
    properties: SourceProperties
    sources: List[SourceData]


# Audio
class AudioProperties(TypedDict, total=False):
    max: int
    min: int
    step: int
    supportedPersistence: bool


class Audio(TypedDict, total=False):
    persistence: bool
    properties: Optional[AudioProperties]
    value: int


# Accessories
class AccessoryData(TypedDict, total=False):
    available: bool
    configurationStatus: str
    serialnum: str
    type: str
    version: str
    wireless: bool


class Accessories(TypedDict, total=False):
    controllable: Optional[dict]
    enabled: Optional[dict]
    pairing: bool
    rears: Optional[List[AccessoryData]]
    subs: Optional[List[AccessoryData]]


# Battery
class Battery(TypedDict, total=False):
    chargeStatus: str
    chargerConnected: str
    minutesToEmpty: int
    minutesToFull: int
    percent: int
    sufficientChargerConnected: bool
    temperatureState: str


# AudioMode
class AudioModeProperties(TypedDict, total=False):
    supportedPersistence: List[str]
    supportedValues: List[str]


class AudioMode(TypedDict):
    persistence: str
    properties: AudioModeProperties
    value: str


# Dual Mono Settings
class DualMonoSettingsProperties(TypedDict, total=False):
    supportedValues: List[str]


class DualMonoSettings(TypedDict):
    value: str
    properties: DualMonoSettingsProperties


# Rebroadcast Latency Mode
class RebroadcastLatencyModeProperties(TypedDict, total=False):
    supportedModes: List[str]


class RebroadcastLatencyMode(TypedDict):
    mode: str
    properties: RebroadcastLatencyModeProperties


# Active Groups
class ProductStateEnum(enumerate):
    WAITING = "WAITING"
    MASTER = "MASTER"


class ProductState(TypedDict):
    productId: str
    status: ProductStateEnum


class ProductRoleEnum(enumerate):
    NORMAL = "NORMAL"


class Product(TypedDict):
    productId: str
    role: ProductRoleEnum


class ActiveGroup(TypedDict):
    activeGroupId: str
    groupMasterId: str
    # groups: List[???]
    name: str
    productStates: List[ProductState]
    products: List[Product]


# System timeout
class SystemTimeout(TypedDict):
    noAudio: bool
    noVideo: bool


# CEC Settings
class CecSettingsSupportedValuesEnum(enumerate):
    ON = "ON"
    OFF = "OFF"
    ALTERNATE_ON = "ALTERNATE_ON"
    ALTMODE_3 = "ALTMODE_3"
    ALTMODE_4 = "ALTMODE_4"
    ALTMODE_5 = "ALTMODE_5"
    ALTMODE_6 = "ALTMODE_6"
    ALTMODE_7 = "ALTMODE_7"


class CecSettingsProperties(TypedDict, total=False):
    supportedModes: List[CecSettingsSupportedValuesEnum]


class CecSettings(TypedDict):
    mode: CecSettingsSupportedValuesEnum
    properties: CecSettingsProperties


# Product Settings
class PresetMetadata(TypedDict):
    accountID: str
    image: str
    name: str
    subType: str


class ContentItem(TypedDict):
    containerArt: str
    location: str
    name: str
    presetable: bool
    source: str
    sourceAccount: str
    type: str


class PresetActionPayload(TypedDict):
    contentItem: ContentItem


class PresetAction(TypedDict):
    actionType: str
    metadata: PresetMetadata
    payload: PresetActionPayload


class Preset(TypedDict):
    actions: List[PresetAction]


class Presets(TypedDict):
    presets: Dict[str, Preset]


class ProductProperties(TypedDict):
    supportedLanguages: List[str]


class ProductSettings(TypedDict):
    language: str
    ntpSyncDone: bool
    presets: Presets
    productName: str
    properties: ProductProperties
    timeformat: str
    timezone: str


class NetworkStateEnum(enumerate):
    DOWN = "DOWN"
    UP = "UP"


class NetworkTypeEnum(enumerate):
    WIRED_USB = "WIRED_USB"
    WIRED_ETH = "WIRED_ETH"
    WIRELESS = "WIRELESS"


class IpInfo(TypedDict):
    ipAddress: str
    subnetMask: str


class NetworkInterface(TypedDict):
    ipInfo: Optional[IpInfo]
    macAddress: str
    name: str
    state: NetworkStateEnum
    type: NetworkTypeEnum


class NetworkStatus(TypedDict):
    interfaces: List[NetworkInterface]
    isPrimaryUp: bool
    primary: str
    primaryIpAddress: str



"""
BoseAuth Module

This module provides functionality to obtain a control token from the BOSE online API,
which is used for local control of a Bose speaker. The control token is a JWT with a limited
lifetime and must be refreshed periodically. The API keys used are publicly available on the
BOSE website, so they are not considered sensitive.

Note:
    This API is not officially supported by Bose and was reverse engineered by analyzing
    the Bose app's API calls. Therefore, the API may change or stop working at any time.
    Please be respectful with the requests to avoid being blocked.
"""

import requests
import time
import json
import logging
import sys
import jwt
from typing import TypedDict, Optional, Dict, Any, cast

from .GSSDK import GSRequest, SigUtils
from .BoseCloudResponse import BoseApiProduct

# --- API Types ---


class SocializeSDKConfigResponseIds(TypedDict):
    """
    Represents the 'ids' portion of the response from the Socialize SDK Config endpoint.

    Attributes:
        gmid (str): The GMID value.
        ucid (str): The UCID value.
    """

    gmid: str
    ucid: str


class SocializeSDKConfigResponse(TypedDict, total=False):
    """
    Represents the full response from the Socialize SDK Config endpoint.

    Attributes:
        appIds (Dict[str, Any]): Application IDs (optional).
        callId (str): The call identifier.
        errorCode (int): The error code, if any.
        errorReportRules (list[Any]): Any error report rules.
        ids (SocializeSDKConfigResponseIds): The GMID and UCID values.
        permissions (Dict[str, list[str]]): Permissions by service.
        statusCode (int): HTTP status code.
        statusReason (str): HTTP status reason.
        time (str): Timestamp of the response.
    """

    appIds: Dict[str, Any]
    callId: str
    errorCode: int
    errorReportRules: list[Any]
    ids: SocializeSDKConfigResponseIds
    permissions: Dict[str, list[str]]
    statusCode: int
    statusReason: str
    time: str


class AccountsLoginResponseSessionInfo(TypedDict):
    """
    Represents session information from the accounts.login endpoint.

    Attributes:
        sessionToken (str): The session token.
        sessionSecret (str): The secret used for signing requests.
    """

    sessionToken: str
    sessionSecret: str


class AccountsLoginResponseUserInfo(TypedDict):
    """
    Represents user information from the accounts.login endpoint.

    Attributes:
        UID (str): The unique user ID.
        signatureTimestamp (str): Timestamp of the signature.
        UIDSignature (str): The UID signature.
    """

    UID: str
    signatureTimestamp: str
    UIDSignature: str


class AccountsLoginResponse(TypedDict):
    """
    Represents the complete response from the accounts.login endpoint.

    Attributes:
        sessionInfo (AccountsLoginResponseSessionInfo): Session info.
        userInfo (AccountsLoginResponseUserInfo): User info.
    """

    sessionInfo: AccountsLoginResponseSessionInfo
    userInfo: AccountsLoginResponseUserInfo


class AccountsGetJWTResponse(TypedDict):
    """
    Represents the response from the accounts.getJWT endpoint.

    Attributes:
        apiVersion (int): The API version.
        callId (str): The call identifier.
        errorCode (int): The error code.
        id_token (str): The JWT token.
        statusCode (int): The HTTP status code.
        statusReason (str): The HTTP status reason.
        time (str): The time of the response.
    """

    apiVersion: int
    callId: str
    errorCode: int
    id_token: str
    statusCode: int
    statusReason: str
    time: str


class IDJwtCoreTokenResponse(TypedDict):
    """
    Represents the response from the id.api.bose.io /id-jwt-core/token endpoint.

    Attributes:
        access_token (str): The access token.
        bosePersonID (str): The Bose person ID.
        expires_in (int): Token expiry time in seconds.
        refresh_token (str): The refresh token.
        scope (str): The token scope.
        token_type (str): The token type.
    """

    access_token: str
    bosePersonID: str
    expires_in: int
    refresh_token: str
    scope: str
    token_type: str


class UsersApiBoseProductResponse(TypedDict, total=False):
    """
    Represents the response from the users.api.bose.io /passport-core/products/ endpoint.

    Attributes:
        attributes (Dict[str, Any]): Product attributes.
        createdOn (str): Creation timestamp.
        groups (list[Any]): Groups information.
        persons (Dict[str, str]): Mapping of person IDs to roles.
        presets (Dict[str, Any]): Presets information.
        productColor (int): The color code of the product.
        productID (str): The product identifier.
        productType (str): The product type.
        serviceAccounts (list[Dict[str, Any]]): Service account details.
        settings (Dict[str, Any]): Additional settings.
        updatedOn (str): Last updated timestamp.
        users (Dict[str, Any]): Users associated with the product.
    """

    attributes: Dict[str, Any]
    createdOn: str
    groups: list[Any]
    persons: Dict[str, str]
    presets: Dict[str, Any]
    productColor: int
    productID: str
    productType: str
    serviceAccounts: list[Dict[str, Any]]
    settings: Dict[str, Any]
    updatedOn: str
    users: Dict[str, Any]


# --- Internal Types ---


class ControlToken(TypedDict):
    """
    Represents a control token with associated information.

    Attributes:
        access_token (str): The access token.
        refresh_token (str): The refresh token.
        bose_person_id (str): The Bose person identifier.
    """

    access_token: str
    refresh_token: str
    bose_person_id: str


# Internal raw token type, based on IDJwtCoreTokenResponse.
RawControlToken = IDJwtCoreTokenResponse


class LoginResponse(TypedDict):
    """
    Represents an internal structure for login responses.

    Attributes:
        session_token (str): The session token.
        session_secret (str): The session secret.
        uid (str): The user ID.
        signatureTimestamp (str): Timestamp for the signature.
        UIDSignature (str): The UID signature.
    """

    session_token: str
    session_secret: str
    uid: str
    signatureTimestamp: str
    UIDSignature: str


# --- BoseAuth Class ---


class BoseAuth:
    """
    A class to interact with the BOSE online API for obtaining control tokens.

    This class uses publicly available API keys to obtain a JWT control token, which is used
    to control a local Bose speaker. It also provides methods to refresh tokens and to fetch
    product information from the BOSE API.

    Attributes:
        GIGYA_API_KEY (str): Public API key for Gigya.
        GIGYA_UA (str): User-Agent string used for Gigya requests.
        BOSE_API_KEY (str): Public API key for the BOSE API.
    """

    GIGYA_API_KEY: str = (
        "3_7PoVX7ELjlWyppFZFGia1Wf1rNGZv_mqVgtqVmYl3Js-hQxZiFIU8uHxd8G6PyNz"
    )
    GIGYA_UA: str = "Bose/32768 MySSID/1568.300.101 Darwin/24.2.0"
    BOSE_API_KEY: str = "67616C617061676F732D70726F642D6D61647269642D696F73"

    def __init__(self) -> None:
        """
        Initialize a new BoseAuth instance.

        The control token, email, and password are initially unset.
        """
        self._control_token: Optional[RawControlToken] = None
        self._email: Optional[str] = None
        self._password: Optional[str] = None

    def set_access_token(
        self, access_token: str, refresh_token: str, bose_person_id
    ) -> None:
        """
        Set the access token and refresh token.

        Args:
            access_token (str): The access token.
            refresh_token (str): The refresh token.
        """
        self._control_token = {
            "access_token": access_token,
            "refresh_token": refresh_token,
            "bosePersonID": bose_person_id,
        }

    def _get_ids(self) -> Optional[Dict[str, str]]:
        """
        Start a session and retrieve the GMID and UCID via the Socialize SDK Config endpoint.

        Returns:
            Optional[Dict[str, str]]: A dictionary with 'gmid' and 'ucid' keys if successful; otherwise, None.
        """
        logging.debug("Getting GMID and UCID")
        url: str = "https://socialize.us1.gigya.com/socialize.getSDKConfig"
        data: Dict[str, Any] = {
            "apikey": self.GIGYA_API_KEY,
            "format": "json",
            "httpStatusCodes": False,
            "include": "permissions,ids,appIds",
            "sdk": "ios_swift_1.0.8",
            "targetEnv": "mobile",
        }
        try:
            response_json: Dict[str, Any] = requests.post(url, data=data).json()
            config: SocializeSDKConfigResponse = cast(
                SocializeSDKConfigResponse, response_json
            )
        except Exception as e:
            logging.error(f"Error getting GMID and UCID: {e}")
            return None

        logging.debug(f"_get_ids: {json.dumps(config, indent=4)}")
        ids = config.get("ids")
        if ids and "gmid" in ids and "ucid" in ids:
            return {"gmid": ids["gmid"], "ucid": ids["ucid"]}
        return None

    def _login(self, email: str, password: str, gmid: str, ucid: str) -> LoginResponse:
        """
        Perform login to Gigya using the provided email and password.

        Args:
            email (str): The user's email address.
            password (str): The user's password.
            gmid (str): The GMID retrieved from _get_ids.
            ucid (str): The UCID retrieved from _get_ids.

        Returns:
            LoginResponse: A dictionary containing session token, session secret, UID, signature timestamp, and UID signature.

        Raises:
            ValueError: If the login fails.
        """
        logging.debug(f"Logging in with {email}, gmid {gmid}, ucid {ucid}")
        url: str = "https://accounts.us1.gigya.com/accounts.login"
        headers: Dict[str, str] = {
            "Host": "accounts.us1.gigya.com",
            "Connection": "keep-alive",
            "Accept": "*/*",
            "User-Agent": self.GIGYA_UA,
            "Accept-Language": "de-DE,de;q=0.9",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data: Dict[str, Any] = {
            "apikey": self.GIGYA_API_KEY,
            "format": "json",
            "gmid": gmid,
            "httpStatusCodes": "false",
            "include": "profile,data,emails,subscriptions,preferences,",
            "includeUserInfo": "true",
            "lang": "de",
            "loginID": email,
            "loginMode": "standard",
            "password": password,
            "sdk": "ios_swift_1.0.8",
            "sessionExpiration": "0",
            "source": "showScreenSet",
            "targetEnv": "mobile",
            "ucid": ucid,
        }
        response_obj: requests.Response = requests.post(url, headers=headers, data=data)
        if response_obj.status_code == 200:
            json_response: Dict[str, Any] = response_obj.json()
            logging.debug(
                "WARNING! CONFIDENTIAL INFORMATION! REMOVE AT LEAST THE session_secret AND UIDSignature FROM THE LOGS!"
            )
            logging.debug(f"_login: {json.dumps(json_response, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
            login_resp: AccountsLoginResponse = cast(
                AccountsLoginResponse, json_response
            )
            return {
                "session_token": login_resp["sessionInfo"]["sessionToken"],
                "session_secret": login_resp["sessionInfo"]["sessionSecret"],
                "uid": login_resp["userInfo"]["UID"],
                "signatureTimestamp": login_resp["userInfo"]["signatureTimestamp"],
                "UIDSignature": login_resp["userInfo"]["UIDSignature"],
            }
        else:
            raise ValueError(f"Login failed: {response_obj.text}")

    def _get_jwt(self, user: LoginResponse, gmid: str, ucid: str) -> Optional[str]:
        """
        Retrieve a JWT token from Gigya using the accounts.getJWT endpoint.

        Args:
            user (LoginResponse): The login response obtained from _login.
            gmid (str): The GMID value.
            ucid (str): The UCID value.

        Returns:
            Optional[str]: The JWT token if successful; otherwise, None.
        """
        url: str = "https://accounts.us1.gigya.com/accounts.getJWT"
        headers: Dict[str, str] = {
            "Host": "accounts.us1.gigya.com",
            "Connection": "keep-alive",
            "Accept": "*/*",
            "User-Agent": self.GIGYA_UA,
            "Accept-Language": "de-DE,de;q=0.9",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        timestamp: str = str(int(time.time()))
        params: Dict[str, Any] = {
            "apikey": self.GIGYA_API_KEY,
            "format": "json",
            "gmid": gmid,
            "httpStatusCodes": "false",
            "nonce": f"{timestamp}_1637928129",
            "oauth_token": user["session_token"],
            "sdk": "ios_swift_1.0.8",
            "targetEnv": "mobile",
            "timestamp": timestamp,
            "ucid": ucid,
        }

        request = GSRequest()
        base_string: str = request.calcOAuth1BaseString("POST", url, True, params)
        sig: str = SigUtils.calcSignature(base_string, user["session_secret"])
        params["sig"] = sig

        try:
            logging.debug("WAARNING! CONFIDENTIAL INFORMATION!")
            response_json: Dict[str, Any] = requests.post(
                url, headers=headers, data=params
            ).json()
            jwt_resp: AccountsGetJWTResponse = cast(
                AccountsGetJWTResponse, response_json
            )
            logging.debug(f"_get_jwt: {json.dumps(jwt_resp, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
        except Exception as e:
            logging.error(f"Error getting JWT: {e}")
            return None
        return jwt_resp.get("id_token")

    def do_token_refresh(
        self, access_token: Optional[str] = None, refresh_token: Optional[str] = None
    ) -> ControlToken:
        """
        Refresh the control token using the id.api.bose.io endpoint.

        If access_token and refresh_token are not provided, the previously stored tokens are used.

        Args:
            access_token (Optional[str]): Existing access token.
            refresh_token (Optional[str]): Existing refresh token.

        Returns:
            ControlToken: A dictionary containing the new access token, refresh token, and Bose person ID.

        Raises:
            ValueError: If no control token is stored or required tokens are missing.
        """
        if self._control_token is None:
            raise ValueError("No control token stored to refresh.")
        if access_token is None:
            access_token = self._control_token.get("access_token")
        if refresh_token is None:
            refresh_token = self._control_token.get("refresh_token")

        if access_token is None or refresh_token is None:
            raise ValueError(
                "Provide both the access_token and refresh_token or the control token"
            )

        fetched: Optional[RawControlToken] = self._fetch_keys(
            access_token=access_token, refresh_token=refresh_token
        )
        if fetched is None:
            raise ValueError("Failed to refresh token")
        self._control_token = fetched
        return {
            "access_token": self._control_token.get("access_token"),
            "refresh_token": self._control_token.get("refresh_token"),
            "bose_person_id": self._control_token.get("bosePersonID", ""),
        }

    def _fetch_keys(
        self,
        gigya_jwt: Optional[str] = None,
        signature_timestamp: Optional[str] = None,
        uid: Optional[str] = None,
        uid_signature: Optional[str] = None,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ) -> Optional[RawControlToken]:
        """
        Fetch the local control token from the id.api.bose.io endpoint.

        Either provide:
            - gigya_jwt, signature_timestamp, uid, and uid_signature, or
            - access_token and refresh_token.

        Args:
            gigya_jwt (Optional[str]): JWT token from Gigya.
            signature_timestamp (Optional[str]): Signature timestamp.
            uid (Optional[str]): User ID.
            uid_signature (Optional[str]): UID signature.
            access_token (Optional[str]): Existing access token.
            refresh_token (Optional[str]): Existing refresh token.

        Returns:
            Optional[RawControlToken]: The fetched token if successful, otherwise None.
        """
        if (
            gigya_jwt is None
            or signature_timestamp is None
            or uid is None
            or uid_signature is None
        ) and (access_token is None or refresh_token is None):
            raise ValueError(
                "Provide either the gigya_jwt, signature_timestamp, uid and uid_signature or the access_token and refresh_token"
            )

        url: str = "https://id.api.bose.io/id-jwt-core/token"
        headers: Dict[str, str] = {
            "X-ApiKey": self.BOSE_API_KEY,
            "X-Software-Version": "10.6.6-32768",
            "X-Api-Version": "1",
            "User-Agent": "MadridApp/10.6.6 (com.bose.bosemusic; build:32768; iOS 18.3.0) Alamofire/5.6.2",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
        }

        if access_token is not None and refresh_token is not None:
            data: Dict[str, Any] = {
                "scope": "openid",
                "client_id": self.BOSE_API_KEY,
                "grant_type": "refresh_token",
                "refresh_token": refresh_token,
            }
        else:
            data = {
                "id_token": gigya_jwt,
                "scope": "openid",
                "grant_type": "id_token",
                "signature_timestamp": signature_timestamp,
                "uid_signature": uid_signature,
                "uid": uid,
                "client_id": self.BOSE_API_KEY,
            }

        try:
            response_json: Dict[str, Any] = requests.post(
                url, headers=headers, json=data
            ).json()
            logging.debug("WARNING! CONFIDENTIAL INFORMATION!")
            logging.debug(f"_fetch_keys: {json.dumps(response_json, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
        except Exception as e:
            logging.error(f"Error fetching keys: {e}")
            return None
        token_resp: IDJwtCoreTokenResponse = cast(IDJwtCoreTokenResponse, response_json)
        return token_resp

    def get_token_validity_time(self, token: str = None) -> int:
        """
        Get the validity time of the given token.

        Args:
            token (str): The JWT token to check.

        Returns:
            int: The time until the token expires in seconds.
        """

        if token is None:
            token = self._control_token.get("access_token")
        if token is None:
            return 0

        try:
            decoded: Dict[str, Any] = jwt.decode(
                token, options={"verify_signature": False}
            )
            exp: int = decoded.get("exp", 0)
            return exp - int(time.time())
        except Exception as e:
            logging.error(f"Error decoding token: {e}")
            return 0

    def is_token_valid(self, token: str = None) -> bool:
        """
        Check if the given token is still valid by decoding it without verifying the signature.

        Args:
            token (str): The JWT token to validate.

        Returns:
            bool: True if the token has not expired, False otherwise.
        """

        if token is None:
            token = self._control_token.get("access_token")
        if token is None:
            return False

        try:
            decoded: Dict[str, Any] = jwt.decode(
                token, options={"verify_signature": False}
            )
            exp: int = decoded.get("exp", 0)
            valid: bool = exp > int(time.time())
            if self._control_token is None:
                self._control_token = {"access_token": token}
            else:
                self._control_token["access_token"] = token
            return valid
        except Exception:
            return False

    def getControlToken(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        forceNew: bool = False,
    ) -> ControlToken:
        """
        Obtain the control token for accessing the local speaker API.

        If a token is already stored and valid, it is returned (unless forceNew is True). Otherwise, the
        token is retrieved by logging in and fetching keys from the Bose API.

        Args:
            email (Optional[str]): User's email address.
            password (Optional[str]): User's password.
            forceNew (bool): If True, force retrieval of a new token even if one is stored.

        Returns:
            ControlToken: A dictionary containing the access token, refresh token, and Bose person ID.

        Raises:
            ValueError: If email and password are not provided for the initial call or if token retrieval fails.
        """
        if not forceNew and self._control_token is not None:
            access_token: Optional[str] = self._control_token.get("access_token")
            if access_token and self.is_token_valid(access_token):
                return {
                    "access_token": access_token,
                    "refresh_token": self._control_token.get("refresh_token", ""),
                }
            else:
                logging.debug("Token is expired. Trying to refresh token")

        if email is not None:
            self._email = email
        if password is not None:
            self._password = password

        if self._email is None or self._password is None:
            raise ValueError("Email and password are required for the first call!")

        ids = self._get_ids()
        if ids is None:
            raise ValueError("Could not retrieve GMID and UCID")
        gmid: str = ids["gmid"]
        ucid: str = ids["ucid"]

        user: LoginResponse = self._login(self._email, self._password, gmid, ucid)
        gigya_jwt: Optional[str] = self._get_jwt(user, gmid, ucid)
        if gigya_jwt is None:
            raise ValueError("Failed to retrieve Gigya JWT")

        fetched: Optional[RawControlToken] = self._fetch_keys(
            gigya_jwt, user["signatureTimestamp"], user["uid"], user["UIDSignature"]
        )
        if fetched is None:
            raise ValueError("Failed to fetch control token")
        self._control_token = fetched
        return {
            "access_token": self._control_token.get("access_token"),
            "refresh_token": self._control_token.get("refresh_token", ""),
            "bose_person_id": self._control_token.get("bosePersonID", ""),
        }

    def fetchProductInformation(self, gwid: str) -> Optional[BoseApiProduct]:
        """
        Fetch product information from the users.api.bose.io endpoint.

        Args:
            gwid (str): The product (or device) identifier.

        Returns:
            Optional[BoseApiProduct]: An instance of BoseApiProduct populated with the response data,
            or None if the fetch fails.
        """
        url: str = f"https://users.api.bose.io/passport-core/products/{gwid}"
        headers: Dict[str, str] = {
            "X-ApiKey": self.BOSE_API_KEY,
            "X-Software-Version": "10.6.6-32768",
            "X-Api-Version": "1",
            "User-Agent": "MadridApp/10.6.6 (com.bose.bosemusic; build:32768; iOS 18.3.0) Alamofire/5.6.2",
            "X-User-Token": self._control_token.get("access_token")
            if self._control_token
            else "",
        }
        try:
            response_json: Dict[str, Any] = requests.get(url, headers=headers).json()
            logging.debug(f"product info: {json.dumps(response_json, indent=4)}")
        except Exception as e:
            logging.error(f"Error fetching product information: {e}")
            return None
        product_resp: UsersApiBoseProductResponse = cast(
            UsersApiBoseProductResponse, response_json
        )
        return BoseApiProduct(**product_resp)



#! /usr/bin/env python
# -*- encoding: utf-8 -*-
"""
Slighly modified version of the Gigya SDK for Python
https://github.com/SAP/gigya-python-sdk/tree/main
"""
from __future__ import unicode_literals
import sys

MAJOR_VERSION_INDEX = 0
MINOR_VERSION_INDEX = 1

PY_3 = sys.version_info[MAJOR_VERSION_INDEX] >= 3

PY_3_9_OR_GREATER = PY_3 and sys.version_info[MINOR_VERSION_INDEX] >= 9

if PY_3:
    import urllib.request
    from urllib.request import build_opener
    from urllib.request import HTTPHandler
    from urllib.request import ProxyHandler
    from urllib.request import install_opener
    from urllib.request import urlopen
    from http.client import HTTPSConnection
    from http.client import HTTPS_PORT
    from urllib.request import HTTPSHandler
    from urllib.parse import quote_plus
    from urllib.parse import urlparse
else:
    import urllib2
    from urllib2 import build_opener
    from urllib2 import HTTPHandler
    from urllib2 import ProxyHandler
    from urllib2 import install_opener
    from urllib2 import urlopen
    from httplib import HTTPSConnection
    from httplib import HTTPS_PORT
    from urllib2 import HTTPSHandler
    from urllib import quote_plus
    from urlparse import urlparse

import hmac
import time
import calendar
import socket
import os
import ssl
import copy

from hashlib import sha1
from base64 import b64decode, b64encode
from json import loads
from json import dumps as jsonstringify
from re import search
from random import randrange

if PY_3:
    string_types = str
    integer_types = int
else:
    string_types = basestring
    integer_types = (int, long)


class GSException(Exception):
    def __init__(self, value):
        self.errorMessage = value

    def __str__(self):
        return repr(self.value)


class GSRequest():
    DEFAULT_API_DOMAIN = "us1.gigya.com"
    VERSION = "3.5.2"
    caCertsPath = os.path.join(os.path.dirname(__file__), "cacert.pem")

    _domain = "accounts.us1.gigya.com"
    _path = ""
    _traceLog = []
    _method = ""
    _proxy = ""
    _host = ""

    _apiKey = ""
    _secretKey = ""
    _userKey = ""
    _params = {}
    _useHTTPS = False
    _apiDomain = DEFAULT_API_DOMAIN

    def __init__(self, apiKey=None, secretKey=None, apiMethod=None, params=None, useHTTPS=False, userKey=None, enable_host_check=True):
        """
         Constructs a request using the apiKey and secretKey.
         @param apiKey
         @param secretKey
         @param apiMethod the api method (including namespace) to call. for example: socialize.getUserInfo
         If namespaces is not supplied "socialize" is assumed
         @param params the request parameters
         @param useHTTPS useHTTPS set this to true if you want to use HTTPS.
         @param userKey A key of an admin user with extra permissions.
         @param enable_host_check Hostcheck can be disabled by passing enable_host_check=false when using HTTPS. Can be necessary when using proxy. Default is True
         If this parameter is provided, then the secretKey parameter is assumed to be the admin user's secret key and not the site's secret key.
        """

        if apiMethod is None:
            return

        self._method = apiMethod

        if params is None:
            self._params = {}
        elif isinstance(params, dict):
            self._params = copy.copy(params)
        elif isinstance(params, string_types):
            self._params = Utils.jsonparse(params)
        else:
            self._params = dict([k, v.encode('utf-8')] for k, v in params.items())

        self._domain = self._params.get("_host", self._domain)
        self._useHTTPS = useHTTPS
        self._apiKey = apiKey
        self._secretKey = secretKey
        self._userKey = userKey
        self._traceLog = list()
        self._enableHostCheck = enable_host_check
        self.traceField("apiMethod", apiMethod)
        self.traceField("apiKey", apiKey)

    def setParam(self, param, val):
        self._params[param] = val

    def getParams(self):
        return self._params

    def setAPIDomain(self, apiDomain=None):
        if apiDomain is None:
            self._apiDomain = self.DEFAULT_API_DOMAIN
        else:
            self._apiDomain = apiDomain

    def setProxy(self, proxy):
        self._proxy = proxy
        self.traceField("proxy", proxy)

    def send(self, timeout=None):
        """Send the request synchronously"""

        if self._method[0:1] == "/":
            self._method = self._method[1:]

        if self._method.find(".") == -1:
            self._domain = "socialize." + self._apiDomain
            self._path = "/socialize." + self._method
        else:
            tokens = self._method.split(".")
            self._domain = tokens[0] + "." + self._apiDomain
            self._path = "/" + self._method

        format = self._params.get("format", None)

        if format is None:
            format = "json"
            self.setParam("format", format)
        if timeout:
            timeout = float(timeout)
            self.traceField("timeout", timeout)

        if (self._method is None or (self._apiKey is None and self._userKey is None)):
            return GSResponse(self._method, None, self._params, 400002, None, self._traceLog)

        if (self._useHTTPS and not os.path.isfile(GSRequest.caCertsPath)):
            return GSResponse(self._method, None, self._params, 400003, None, self._traceLog)

        try:
            self.setParam("httpStatusCodes", "false")
            self.traceField("apiKey", self._apiKey)
            self.traceField("apiMethod", self._method)
            self.traceField("params", self._params)
            self.traceField("useHTTPS", self._useHTTPS)
            self.traceField("userKey", self._userKey)
            responseStr = self.sendRequest("POST", self._host, self._path, self._params, self._apiKey, self._secretKey, self._useHTTPS, timeout, self._userKey, self._enableHostCheck)

            return GSResponse(self._method, responseStr, None, 0, None, self._traceLog)

        except Exception as ex:
            errCode = 500000
            errMsg = str(ex)
            length = len("Operation timed out")

            if (errMsg[0:length] == "Operation timed out"):
                errCode = 504002
                errMsg = "Request Timeout"

            return GSResponse(self._method, None, self._params, errCode, errMsg, self._traceLog)

    def sendRequest(self, httpMethod, domain, path, params, token, secret=None, useHTTPS=False, timeout=None, userKey=None, enable_host_check=True):

        params["sdk"] = "ios_swift_1.0.8"
        # prepare query params
        protocol = "https" if (useHTTPS or not secret) else "http"
        resourceURI = protocol + "://" + self._domain + path

        timestamp = calendar.timegm(time.gmtime())

        # unique token
        nonce = str(SigUtils.currentTimeMillis()) + str(randrange(1000))

        if userKey:
            params["userKey"] = userKey

        if secret:
            params["apiKey"] = token
            params["timestamp"] = timestamp
            params["nonce"] = nonce

            # signature
            signature = self.getOAuth1Signature(secret, httpMethod, resourceURI, useHTTPS, params)
            params["sig"] = signature
        else:
            params["oauth_token"] = token

        # get rest response.
        res = self.curl(resourceURI, params, timeout, enable_host_check)

        return res


    def curl(self, url, params=None, timeout=None, enable_host_check=True):

        queryString = self.buildQS(params)

        self.traceField("URL", url)
        self.traceField("postData", queryString)

        proto = "https" if self._useHTTPS else "http"

        if self._proxy:
            opener = build_opener(
                HTTPHandler(),
                ValidHTTPSHandler(enable_host_check), # maybe always send False???
                ProxyHandler({proto: self._proxy}))
        else:
            opener = build_opener(
                HTTPHandler(),
                ValidHTTPSHandler(enable_host_check))

        queryString = queryString.encode('utf-8')

        currentOpener = urllib.request._opener if PY_3 else urllib2._opener
        install_opener(opener)

        if timeout:
            response = urlopen(url, queryString, timeout)
        else:
            response = urlopen(url, queryString, context=ssl._create_unverified_context())

        install_opener(currentOpener)

        result = response.read()
        result = result.decode('utf-8')

        return result

    def buildQS(self, params):
        """Converts a params dictionary to a sorted query string"""
        queryString = ""
        amp = ""
        # keys = params.keys()
        keys = list(params.keys())
        keys.sort()
        for key in keys:
            value = params.get(key)
            if value is not None:
                queryString += amp + key + "=" + self.UrlEncode(value)
                amp = "&"

        return queryString

    def getOAuth1Signature(self, key, httpMethod, url, isSecureConnection, requestParams):
        # Create the BaseString.
        baseString = self.calcOAuth1BaseString(httpMethod, url, isSecureConnection, requestParams)
        self.traceField("baseString", baseString)

        return SigUtils.calcSignature(baseString, key)

    def calcOAuth1BaseString(self, httpMethod, url, isSecureConnection, requestParams):
        normalizedUrl = ""
        u = urlparse(url)

        if isSecureConnection:
            protocol = "https"
        else:
            protocol = u.scheme.lower()

        port = u.port

        normalizedUrl += protocol + "://"
        normalizedUrl += u.hostname.lower()

        if port != None and ((protocol == "http" and port != 80) or (protocol == "https" and port != 443)):
            normalizedUrl += ':' + port

        normalizedUrl += u.path

        # Create a sorted list of query parameters
        queryString = self.buildQS(requestParams)

        # Construct the base string from the HTTP method, the URL and the parameters
        baseString = httpMethod.upper() + '&' + self.UrlEncode(normalizedUrl) + '&' + self.UrlEncode(queryString)

        return baseString

    def UrlEncode(self, value):
        if value is None:
            return value
        elif isinstance(value, integer_types):
            return str(value)
        else:
            if isinstance(value, dict) or isinstance(value, list):
                str_value = jsonstringify(value)
            else:
                str_value = value.encode('utf-8')

            return quote_plus(str_value).replace("+", "%20").replace("%7E", "~")

    def traceField(self, name, value):
        if value:
            self._traceLog.append(str(name) + "=" + repr(value))

    @staticmethod
    def setCACertsPath(path):
        GSRequest.caCertsPath = path


class GSResponse():
    """
     Wraps the server's response.
     If the request was sent with the format set to "json", the getData() will return null and you should use getResponseText() instead.
     We only parse response text into a dictionary if request format is set "json."
    """

    errorCode = 0
    errorMessage = None
    rawData = ""
    data = {}
    params = None
    method = None
    traceLog = []
    errorMsgDic = {}
    errorMsgDic[400002] = "Required parameter is missing"
    errorMsgDic[500000] = "General server error"

    def getErrorCode(self):
        return self.errorCode

    def getErrorMessage(self):
        if self.errorMessage:
            return self.errorMessage
        else:
            if (self.errorCode == 0) or (not self.errorMsgDic.get(self.errorCode)):
                return ""
            else:
                return self.errorMsgDic.get(self.errorCode)

    def getResponseText(self):
        if (type(self.rawData) == "bytes"):
            self.rawData = self.rawData.decode(encoding="utf-8", errors="strict")
        return self.rawData

    def getData(self):
        if (type(self.data) == "bytes"):
            self.data = self.data.decode(encoding="utf-8", errors="strict")
        return self.data

    def getObject(self, key):
        return self.data.get(key)

    def traceField(self, name, value):
        if value:
            self.traceLog.append(str(name) + "=" + repr(value))

    def __init__(self, method, responseText=None, params=None, errorCode=None, errorMessage=None, traceLog=None):
        self.traceLog = traceLog
        self.method = method

        if params:
            self.params = params
        else:
            self.params = {}

        if responseText:
            self.traceField("responseText", responseText)

            self.rawData = responseText.encode('utf-8')

            if (responseText.lstrip().find('{') != -1):
                self.data = Utils.jsonparse(responseText)

                if self.data:
                    self.errorCode = self.data.get("errorCode")
                    self.errorMessage = self.data.get("errorMessage")

            else:
                matches = search(r"<errorCode>([^>]+)</errorCode>", self.rawData)
                if matches:
                    errCodeStr = matches.group(1)
                    if errCodeStr:
                        self.errorCode = int(errCodeStr)
                        matches = search(r"<errorMessage>([^>]+)</errorMessage>", self.rawData)
                        if matches:
                            self.errorMessage = matches.group(1)

        else:
            self.errorCode = errorCode
            self.errorMessage = errorMessage if errorMessage != None else self.getErrorMessage()
            self.populateClientResponseText()
            self.traceField("responseText", self.rawData)

    def populateClientResponseText(self):
        if self.params.get("format") == 'json':
            self.rawData = "{errorCode:" + str(self.errorCode) + ",errorMessage:\"" + self.errorMessage + "\"}"
        else:
            sb = [
                "<?xml version=\"1.0\" encoding=\"utf-8\"?>",
                "<" + self.method + "Response xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"urn:com:gigya:api http://socialize-api.gigya.com/schema\" xmlns=\"urn:com:gigya:api\">",
                "<errorCode>" + str(self.errorCode) + "</errorCode>",
                "<errorMessage>" + self.errorMessage + "</errorMessage>",
                "</" + self.method + "Response>"
            ]
            self.rawData = "\r\n".join(sb)

    def getLog(self):
        return ", ".join(self.traceLog)

    def __str__(self):
        sb = ""
        sb += "\terrorCode:"
        sb += str(self.errorCode)
        sb += "\n\terrorMessage:"
        sb += self.errorMessage
        sb += "\n\tdata:"
        sb += str(self.data)
        return sb


class ValidHTTPSConnection(HTTPSConnection):
    default_port = HTTPS_PORT

    def __init__(self, *args, **kwargs):
        HTTPSConnection.__init__(self, *args, **kwargs)
        self._context = kwargs.get('context')

    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout, self.source_address)

        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        self.sock = self._context.wrap_socket(sock, server_hostname=self.host)


class ValidHTTPSHandler(HTTPSHandler):
    def __init__(self, enable_host_check=True):
        super().__init__()
        self._enableHostCheck = enable_host_check

    def https_open(self, req):
        return self.do_open(self.get_connection, req)

    def get_connection(self, host, timeout):
        return ValidHTTPSConnection(host, timeout = timeout, context = self.create_context())

    def create_context(self):
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.minimum_version = ssl.TLSVersion.TLSv1_2
        return context


class SigUtils():

    @staticmethod
    def validateUserSignature(UID, timestamp, secret, signature, expiration=None):
        baseString = timestamp + "_" + UID
        expectedSig = SigUtils.calcSignature(baseString, secret)
        if expiration is None:
            return expectedSig == signature
        else:
            expired = SigUtils.signatureTimestampExpired(timestamp, expiration)
            return not expired and expectedSig == signature

    @staticmethod
    def validateUserSignatureWithExpiration(UID, timestamp, secret, signature, expiration):
        expired = SigUtils.signatureTimestampExpired(timestamp, expiration)
        signatureValidated = SigUtils.validateUserSignature(UID, timestamp, secret, signature)
        return not expired and signatureValidated

    @staticmethod
    def validateFriendSignature(UID, timestamp, friendUID, secret, signature, expiration=None):
        baseString = timestamp + "_" + friendUID + "_" + UID
        expectedSig = SigUtils.calcSignature(baseString, secret)
        if expiration is None:
            return expectedSig == signature
        else:
            expired = SigUtils.signatureTimestampExpired(timestamp, expiration)
            return not expired and expectedSig == signature

    @staticmethod
    def validateFriendSignatureWithExpiration(UID, timestamp, friendUID, secret, signature, expiration):
        expired = SigUtils.signatureTimestampExpired(timestamp, expiration)
        signatureValidated = SigUtils.validateFriendSignature(UID, timestamp, friendUID, secret, signature)
        return not expired and signatureValidated

    @staticmethod
    def getDynamicSessionSignature(gltCookie, timeoutInSeconds, secret):
        expirationTimeUnixMS = calendar.timegm(time.gmtime()) + timeoutInSeconds
        expirationTimeUnixString = str(expirationTimeUnixMS)
        unsignedExpString = gltCookie + "_" + expirationTimeUnixString
        signedExpString = SigUtils.calcSignature(unsignedExpString, secret)
        ret = expirationTimeUnixString + "_" + signedExpString
        return ret

    @staticmethod
    def getDynamicSessionSignatureUserSigned(gltCookie, timeoutInSeconds, userKey, secret):
        expirationTimeUnixMS = calendar.timegm(time.gmtime()) + timeoutInSeconds
        expirationTimeUnixString = str(expirationTimeUnixMS)
        unsignedExpString = gltCookie + "_" + expirationTimeUnixString + "_" + userKey
        signedExpString = SigUtils.calcSignature(unsignedExpString, secret)
        ret = expirationTimeUnixString + "_" + userKey + "_" + signedExpString
        return ret

    @staticmethod
    def calcSignature(baseString, key):
        encodedBase = baseString.encode("utf-8")
        encodedKey = key.encode('utf-8')
        rawHmac = hmac.new(b64decode(encodedKey), encodedBase, sha1).digest()
        signature = b64encode(rawHmac)
        signature = signature.decode('utf-8')

        return signature

    @staticmethod
    def currentTimeMillis():
        return int(round(time.time() * 1000))

    @staticmethod
    def signatureTimestampExpired(signatureTimestampExpired, expiration):
        now = int(round(time.time()))
        timestamp = int(signatureTimestampExpired)
        return abs(now - timestamp) > expiration

class Utils():

    @staticmethod
    def jsonparse(source):
        if PY_3_9_OR_GREATER:
            return loads(source)
        else:
            return loads(source, encoding='utf-8')




import os
from setuptools import setup, find_packages

version = os.getenv("PACKAGE_VERSION", "0.0.0")

setup(
    name="pybose",
    version=version,
    description="An unofficial Python API for controlling Bose soundbars and speakers.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="cavefire",
    author_email="timo@cavefire.net",
    url="https://github.com/cavefire/pybose",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "zeroconf",
        "websockets",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)


from typing import TypedDict, List, Optional, Dict


class V4VInput(TypedDict):
    input: str
    nameID: str


class Attributes(TypedDict, total=False):
    v4vInputs: List[V4VInput]


class PresetMetadata(TypedDict):
    accountID: str
    image: str
    name: str
    subType: str


class ContentItem(TypedDict):
    containerArt: str
    location: str
    name: str
    presetable: bool
    source: str
    sourceAccount: str
    type: str


class PresetPayload(TypedDict):
    contentItem: ContentItem


class PresetAction(TypedDict):
    actionType: str
    metadata: PresetMetadata
    payload: PresetPayload


class Preset(TypedDict):
    actions: List[PresetAction]


class ServiceAccountTokens(TypedDict, total=False):
    refresh_token: str
    refreshToken: str
    tokenType: str


class ServiceAccountAttributes(TypedDict, total=False):
    alexaEnv: str
    region: str
    WuWModel: str
    WuWord: str
    email: str
    language: str
    isDefaultAccount: bool


class ServiceAccount(TypedDict, total=False):
    accountID: str
    accountType: str
    bosePersonID: str
    createdOn: str
    provider: str
    providerAccountID: str
    tokens: ServiceAccountTokens
    updatedOn: str
    attributes: Optional[ServiceAccountAttributes]
    disabled: Optional[bool]
    name: Optional[str]
    productID: Optional[str]


class UserRole(TypedDict):
    role: str
    trustLevel: str


class Settings(TypedDict):
    language: str
    name: str
    sharingMode: str
    timeFormat: str
    timeZone: str


class BoseApiProduct(TypedDict):
    attributes: Attributes
    createdOn: str
    groups: List[str]
    persons: Dict[str, str]
    presets: Dict[str, Preset]
    productColor: int
    productID: str
    productType: str
    serviceAccounts: List[ServiceAccount]
    settings: Settings
    updatedOn: str
    users: Dict[str, UserRole]
